# coding: utf-8

from math import dist, sqrt, ceil

# def dist(r1, r2):
#     return ((r2[0] - r1[0])**2 + (r2[1] - r1[1])**2)**0.5

def get_direction(r1, r2):
    '''
    r1 -- a 2-tuple of floats (chaser position)
    r2 -- a 2-tuple of floats (target position)
    returns a unit 2d vector pointing from r2 to r1
    '''
    # assert not r1 == r2, 'the points coincide -- the direction is not defined'
    if r1[1] == r2[1]: # horisontal alignment
        return (1 if r2[0] >= r1[0] else -1, 0) # (1,0) if points coincide
    if r1[0] == r2[0]: # vertical alignment
        return (0, 1 if r2[1] > r1[1] else -1)
    d = dist(r1, r2)
    return ((r2[0] - r1[0])/d, (r2[1] - r1[1])/d)

def advance_position(r, v, tau=1):
    '''from the current position r advance to the position
    at the next iteration: r -> r + v
    (maybe I do not need this function?)
    '''
    return (r[0] + v[0]*tau, r[1] + v[1]*tau) 

def has_contact(r1, r2, tol=1e-5):
    return dist(r1, r2) < tol
    # return ((r2[0] - r1[0])**2 + (r2[1] - r1[1])**2)**0.5 < tol

def diff(r1, r2):
    return (r1[0]-r2[0], r1[1]-r2[1])

def norm(v):
    assert v != (0, 0), 'zero vector cannot be normalised'
    l = sqrt(v[0]**2 + v[1]**2)
    return (v[0]/l, v[1]/l)

def _dot_prod(r1, r2):
    return r1[0]*r2[0] + r1[1]*r2[1]

def _cross_prod(r1, r2):
    return r1[0]*r2[1] - r1[1]*r2[0]

def min_passing_distance(r1, v1, n1, tau = 1, r0 = (0,0)):
    '''hound's path form r1 with absolute velocity v1 in the
    direction n1 over the time interval in (0,tau) with hare 
    at rest in r0 (as happens in the hare reference frame); 
    what is the minimal separation distance the two can be at?
    '''
    # P is the closest point on the hound path to r0 (hare's position)
    delta_r = diff(r0, r1)
    l = abs(_dot_prod(delta_r, n1))    # distance from P to r1
    l1 = tau * v1                      # length of the full path (before, on and after P0)
    d = abs(_cross_prod(delta_r, n1))/2 # length of the normal from r0 to P
    if l1 >= l: # the hound will pass through the closest point (normal base)
        return d
    else:       # the hound will stop short of the point P
        return sqrt((l-l1)**2 + d**2)

def get_next_chaser_params(chaser_pos, chaser_speed, target_pos, target_speed, tau):
    ''' given the target position and its speed (direction is assumed (1,0)) and
    the chaser position and its absolute speed, return 2-tuple of
     - chaser deirection in the next time interval (of duration tau)
     - chaser position at the end of that time interval 
    '''
    next_target_pos = advance_position(target_pos, (target_speed*tau, 0))
    chaser_dir = get_direction(chaser_pos, next_target_pos)
    chaser_shift = (chaser_dir[0]*chaser_speed*tau, chaser_dir[1]*chaser_speed*tau)
    next_chaser_pos = advance_position(chaser_pos, chaser_shift)
    return (chaser_dir, next_chaser_pos)

def galilean_transfom(v, V):
    '''returns velocity (of the chaser) in the coordinate system
    which moves with (constant) velocity V (of the target)'''
    return (v[0] - V[0], v[1] - V[1])

def chase(target_speed, chaser_speed, chaser_pos, max_steps, catching_dist=1e-5, tau=1):
    '''
    the usual docstring blah...

    The special cases to improve performance and cut corner cases:
    General condition -- the target and chaser are aligned, ie, chaser_pos[1] = 0
    1. chaser is behind at the start AND fast target (target_speed >= chaser_speed):
       False (always)
    2. chaser is behind at the start AND slow target:
       True OR False depending on available time (max_steps)
    3. chaser is ahead at the start:
       True OR False depending on available time
    '''
    if chaser_pos[1] == 0:
        if chaser_pos[0] < 0 and target_speed >= chaser_speed: # slow chaser
            return False
        separation = abs(chaser_pos[0])
        delta_speed = chaser_speed - target_speed if chaser_pos[0] < 0 else chaser_speed + target_speed
        return ceil(separation/delta_speed) <= max_steps
    # generic (non-aligned) case -- condition for T and C having equal velocities not possible
    # and the chaser movement direction (normalised velocity) is defined
    tp = (0, 0) # target starting position 
    V  = target_speed
    tv = (V, 0) # 2-tuple of target velocity
    cp = chaser_pos
    if has_contact(tp, cp, tol=catching_dist):
        return True
    for i in range(0, max_steps + 1):
        n, next_cp = get_next_chaser_params(cp, chaser_speed, tp, V, tau)
        cv = (n[0]*chaser_speed, n[1]*chaser_speed) # chaser velocity
        # transfer into the target reference frame to check the contact during next step
        cv = galilean_transfom(cv, tv) # chaser velocity transform, target is at rest now
        cs = sqrt(cv[0]**2 + cv[1]**2) # chaser speed in moving ref. frame
        # print(f'chaser position {cp} in rest RF')
        cp = (cp[0] - tv[0]*i*tau, cp[1] - tv[1]*i*tau) # chaser coordinate transform, target is at (0,0) now
        # print(f'chaser position {cp} in moving RF')
        if cv[0]**2 + cv[1]**2 == 0: # target and chaser move with the same velocity -- chased is doomed to fail
            # print(f'target and chaser do not move relative to each other, give up')
            return False
        d = min_passing_distance(cp, cs, norm(cv))
        # print(f'min passing distance {d} at step {i}')
        if d < catching_dist:
            return True
        else: # didn't catch, reset for the next step (in the rest ref. frame)
            cp = next_cp # chaser starting position on the next step
            tp = advance_position(tp, tv) # target starting position on the next step
    return False

    
def test_chase_1():
    # generic cases (low precision, should work even if no "catch-on-the-fly" test is used)
    assert not chase(100,10,(10,10),100)
    assert chase(10,11,(10,10),20, catching_dist=0.5)
    assert chase(1.0,1.1,(10,10),50, catching_dist=0.1)
    assert not chase(100,10,(10,10),100)

def test_chase_2():
    # degenerate (boundary, corner) cases
    assert not chase(10, 5, (5,5), 100) # slow chaser too far from the line -- never catches
    assert not chase(5, 5, (-1,0), 100) # slow chaser behind target -- never catches
    assert chase(5, 10, (-10,0), 20) # fast chaser behind target -- will catch given enough steps
    assert not chase(5, 6, (-10,0), 9) # fast chaser behind target -- won't catch since not enough steps
    assert chase(5, 6, (-10,0), 10) # fast chaser behind target -- will catch at the very end

def test_chase_3(): # TESTS THE CATHC IN-FLIGHT, FAILS IF CONTACT IS CHECKED ONLY AT THE END OF TIME INTERVAL
    assert chase(10,10,(10,10),100)     # contact at the arrival
    #x = 1e-4 # disbalance is larger than default catching_dist but fine tuned for in-flight contact
    #assert chase(10,100,(10,100-x),100) # contact happens at the first step in-flight at tau=1-0.01*x
    assert chase(10,100,(10,100-1e-4),100) # contact happens at the first step in-flight at tau=1-0.01*x
    #x = 1e-3 # disbalance is larger than default catching_dist and out of tune for in-flight contact
    #assert not chase(100,10,(100,x),100)# adjust this similarly to 1st to make catch
    assert not chase(100,10,(100,1e-3),100)# adjust this similarly to 1st to make catch
    #x = 1e-6
    #assert chase(100,10+x,(100,10),100) # very fine balance for slow chaser helps
    assert chase(100,10+1e-6,(100,10),100) # very fine balance for slow chaser helps
    assert not chase(100,10.4,(99.25,10),50) # hopeless to catch if not fine tuning
    assert chase(100,10.0375,(99.25,10),5,catching_dist=0.01) # a bit of fine tuning and lower precision do the trick

def test_chase_4(): # TO BE EXCLUDED FROM TEMPLATE
    # Hancheng's test and its variations -- tests handling corner cases of aligned target and chaser
    assert chase(1,0.4,(0.1,0),10) # slow catcher in the line of target run -- will catch passively
    assert not chase(1,0.5,(100,0),20) # slow catcher and fast target runs towrads each other but not enough time
    assert chase(1,0.5,(100,0),67) # slow catcher and fast target run towards each other and time is enough
    assert not chase(1,0.5,(100,0),66) # slow catcher and fast target run towards each other but time is short

def test_chase_5(): # NEWLY ADDED
    assert not chase(100,9.9,(100,10),100) # slow chaser is too far from the target line
    assert chase(100,10,(100,10 + 0.5*1e-5),100) # lucky initial position of an otherwise hopelessly slow chaser
    assert not chase(100,10,(100,10 + 2.5*1e-5),100) # just notch farther away and chase is lost

